#include <algorithm>
#include <concepts>
#include <iostream>
#include <optional>
#include <utility>
#include <vector>

#include <intcode.hpp>
using namespace intcode;

namespace {

template <class R> auto set_params(Machine machine, R const &params) {
  std::vector<Machine> amps;
  for (auto p : params) {
    auto m = machine;
    std::get<Input>(m.step()).pos = p;
    amps.push_back(std::move(m));
  }
  return amps;
}

template <class R>
auto feed(R &&amps, value_type start) -> std::optional<value_type> {
  for (auto &m : amps) {
    auto i = m.step();
    if (auto p = std::get_if<Input>(&i)) {
      p->pos = start;
      start = std::get<Output>(m.step()).val;
    } else {
      return {};
    }
  }
  return {start};
}

auto compute1(Machine machine, std::vector<value_type> const &params)
    -> value_type {
  return *feed(set_params(std::move(machine), params), 0);
}
auto compute2(Machine machine, std::vector<value_type> const &params)
    -> value_type {
  auto amps = set_params(std::move(machine), params);
  value_type last = 0;
  while (auto next = feed(amps, last)) {
    last = *next;
  }
  return last;
}

template <std::invocable<Machine, std::vector<value_type> const &> F>
auto optimize(Machine machine, std::vector<value_type> params, F f) {
  value_type best = 0;
  do {
    best = std::max(best, f(machine, params));
  } while (std::next_permutation(params.begin(), params.end()));
  return best;
}

} // namespace

auto main() -> int {
  auto machine = Machine{parse_stream(std::cin)};
  std::cout << "Part 1: " << optimize(machine,            {0, 1, 2, 3, 4}, compute1) << std::endl;
  std::cout << "Part 2: " << optimize(std::move(machine), {5, 6, 7, 8, 9}, compute2) << std::endl;
}
